/* free-list.c */

/* fonctions de construction et d'acces a la free-list */


#include "mlvalues.h"
#include "free_list.h"
#include "mem_misc.h"


/* variables globales */
address next_block, free_list_mark;
address *last_link;
address hp;
address block_end;

void make_free_block (adr, len)
     address adr;
     unsigned long len;
{
  if (len <= MAX_SIZE){         /* ce test est presque tj vrai */
    H_HEADER (adr) = MAKE_HEADER (len - 1, TAG_JUNK);
    return;
  }
  while (len > MAX_SIZE + SIZE_HEADER){
    len -= MAX_SIZE;
    H_HEADER (&H_FIELD (adr, len - SIZE_HEADER)) =
                        MAKE_HEADER (MAX_SIZE - 1, TAG_JUNK);
  }
  if (len <= MAX_SIZE){
    H_HEADER (adr) = MAKE_HEADER (len - 1, TAG_JUNK);
  }else{
    Zassert (len == MAX_SIZE + SIZE_HEADER);
    len -= SIZE_HEADER + 1;
    H_HEADER (&H_FIELD (adr, len - SIZE_HEADER)) =
                       MAKE_HEADER (1, TAG_JUNK);
    Zassert (len >= SIZE_HEADER + 1);
    H_HEADER (adr) = MAKE_HEADER (len - 1, TAG_JUNK);
  }
}


struct {
  header hd;
  header *link;
  long padding;
} sentinel1, sentinel2;

#define SENT1 ((address) &sentinel1)
#define SENT2 ((address) &sentinel2)

static address current;
address fl_next;

static blocksize biggest_free_size;

#ifdef DEBUG
static unsigned long free_words;
#endif

void new_free_list ()
{
  last_link = &LINK (SENT1);
  current = SENT2;
  fl_next = SENT2 + 2;
  biggest_free_size = 0;
#ifdef DEBUG
  free_words = 0;
#endif
}

void add_to_free_list (adr, len)
     address adr;
     unsigned long len;
{
  unsigned long size = fl_next - current;

  Zassert (fl_next != adr);

#ifdef DEBUG
  {
    unsigned long i;
    for (i = 1; i < size; i++)
      current [i] = (value) gc_debug_random ();
  }
  free_words += size;
#endif

  if (size >= 2){
    if (size > biggest_free_size){
      biggest_free_size = size;
    }
    make_free_block (current, size);
    for (;current < fl_next; current += H_SIZE (current) + 1){
      *last_link = current;
      last_link = &LINK (current);
    }
  }else{
    if (size == 1){
      H_HEADER (current) = MAKE_HEADER (0, TAG_JUNK);
    }
  }

  current = adr;
  fl_next = adr + len;
}

/* renvoie la taille du plus grand bloc libre. */
blocksize get_biggest_free_block_size ()
{
  blocksize size;

  size = fl_next - current;

  if (size  <= 2) {
    return 0;
  } else if (size > biggest_free_size) {
    return size;
  } else {
    return biggest_free_size;
  }
}

void close_free_list ()
{
  add_to_free_list (NULL, 0);   /* force l'insertion du dernier bloc */
  Zassert (LINK (SENT1) == SENT2);
  if (last_link == &LINK (SENT2)){
    *last_link = next_block = block_end = hp = NULL;
    return;                   /* la free-list est vide */
  }
  next_block = LINK (SENT2);
  hp = block_end = NULL;	/* valeur bidon */
  if (last_link == &LINK (next_block)){	/* un seul bloc dans la liste */
    *last_link = NULL;
  }else{
    *last_link = next_block;
  }
#ifdef DEBUG
  gc_message (" ", free_words, " free words ");
#endif
  return;
}

void extract_next_block ()
{
  Zassert (*last_link == next_block || *last_link == NULL);
  if (next_block != NULL){
    hp = next_block;
    block_end = next_block + H_SIZE (next_block) + 1;
    next_block = LINK (next_block);
  } /* si la free-list est vide, on ne fait rien. */
}

void put_back_current_block ()
{
  Zassert (last_link != NULL);

  if (hp + SIZE_HEADER < block_end){
    *last_link = hp;
    H_HEADER (hp) = MAKE_HEADER ((block_end - hp) - 1, TAG_JUNK);
    last_link = &LINK (hp);
    *last_link = next_block;
  }else{
    if (hp + SIZE_HEADER == block_end){	/* bloc vide (un header seul) */
      H_HEADER (hp) = MAKE_HEADER (0, TAG_JUNK);
    }
    if (&LINK (next_block) != last_link){
      *last_link = next_block;
    }else{
      *last_link = NULL;
    }
  }
}
