/* unix.c : appels systeme Unix */

#include <fcntl.h>
#include "config.h"
#include "mlvalues.h"
#include "globals.h"
#include "alloc.h"
#include "memory.h"
#include "stacks.h"

static int system_flags[] =
{ O_RDONLY, O_WRONLY, O_RDWR, O_CREAT, O_TRUNC, O_EXCL,
  4, 2, 1, 01, 010, 0100, 7, 0111 };

void unix_init(argc, argv)
     int argc;
     char * argv [];
{
  value v;
  int i;

  v = alloc_string_array(argc, argv);
  MODIFY(&FIELD(global_data, UNIX_COMMAND_LINE), v);
  for (i = MODE_RDONLY; i <= RIGHTS_EVERYBODY; i++)
    FIELD(global_data, i) = MLINT(system_flags[i - MODE_RDONLY]);
}

#ifdef macintosh

char * error_message(err)
     int err;
{
  extern char * strerror();
  
  return strerror(err);
}

#else

char * error_message(err)
     int err;
{
  extern int sys_nerr;
  extern char * sys_errlist [];
  if (err < 0 || err >= sys_nerr)
    return "unknown error";
  else
    return sys_errlist[err];
}
#endif  

value os_error_message(errcode)    /* ML */
     value errcode;
{
  return alloc_string(error_message(CINT(errcode)));
}

extern int errno;

value os_errno()                /* ML */
{
  return MLINT(errno);
}

value unix_exit(status)         /* ML */
     value status;
{
  exit(CINT(status));
  return MLINT (0);
  		/* Not reached -- suppress warnings from smart compilers. */
}

value unix_open(path, flags, mode) /* ML */
     value path, flags, mode;
{
#ifdef macintosh
  return MLINT (open (CSTRING (path), CINT (flags)));
#else
  return MLINT (open (CSTRING (path), CINT (flags), CINT (mode)));
#endif
}

value unix_close(des)           /* ML */
     value des;
{
  return MLINT (close (CINT (des)));
}

value unix_read(d, buf, offset, nbytes) /* ML */
     value nbytes, offset, buf, d;
{
  return MLINT (read (CINT (d), &BYTE(buf, CINT (offset)), CINT (nbytes)));
}

value unix_write(d, buf, offset, nbytes) /* ML */
     value nbytes, offset, buf, d;
{
  return MLINT (write (CINT (d), &BYTE(buf, CINT (offset)), CINT (nbytes)));
}

value unix_lseek(des, offset, whence) /* ML */
     value whence, offset, des;
{
  return
    MLINT (lseek (CINT (des), CINT (offset), CINT (whence)));
}

value unix_unlink(name)         /* ML */
     value name;
{
  return MLINT(unlink(CSTRING(name)));
}

/* Ce qui suit n'est en general defini que sur les vraies Unix boxes. */

#ifndef UNIX

/* The return statements are here to suppress warnings from smart compilers. */

value unix_truncate() { failwith("unix_truncate"); return MLINT (0); }
value unix_stat() { failwith("unix_stat"); return MLINT (0); }
value unix_fstat() { failwith("unix_fstat"); return MLINT (0); }
value unix_link() { failwith("unix_link"); return MLINT (0); }
value unix_pipe() { failwith("unix_pipe"); return MLINT (0); }
value unix_fork() { failwith("unix_fork"); return MLINT (0); }
value unix_execv() { failwith("unix_execv"); return MLINT (0); }
value unix_times() { failwith("unix_times"); return MLINT (0); }

#else /* ! UNIX */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <sys/param.h>

value unix_truncate(path, length) /* ML */
     value path, length;
{
  return MLINT(truncate(CSTRING(path), CINT(length)));
}

static value stat_aux(buf)
     struct stat * buf;
{
  value v;
  v = alloc_tuple(13);
  FIELD (v, 0) = MLINT (buf->st_dev);
  FIELD (v, 1) = MLINT (buf->st_ino);
  FIELD (v, 2) = MLINT (buf->st_mode);
  FIELD (v, 3) = MLINT (buf->st_nlink);
  FIELD (v, 4) = MLINT (buf->st_uid);
  FIELD (v, 5) = MLINT (buf->st_gid);
  FIELD (v, 6) = MLINT (buf->st_rdev);
  FIELD (v, 7) = MLINT (buf->st_size);
  FIELD (v, 8) = MLINT (buf->st_atime);
  FIELD (v, 9) = MLINT (buf->st_mtime);
  FIELD (v, 10) = MLINT (buf->st_ctime);
  FIELD (v, 11) = MLINT (buf->st_blksize);
  FIELD (v, 12) = MLINT (buf->st_blocks);
  return v;
}

value unix_stat(path)           /* ML */
     value path;
{
  struct stat stat_buf;

  if (stat(CSTRING(path), &stat_buf) == -1)
    os_error();
  return stat_aux(&stat_buf);
}

value unix_fstat(fd)           /* ML */
     value fd;
{
  struct stat stat_buf;

  if (fstat(CINT(fd), &stat_buf) == -1)
    os_error();
  return stat_aux(&stat_buf);
}

value unix_link(name1, name2)   /* ML */
     value name1, name2;
{
  return MLINT(link(CSTRING(name1), CSTRING(name2)));
}

value unix_pipe()               /* ML */
{
  int fd[2];
  value res;

  if (pipe(fd) == -1)
    os_error();
  res = alloc_tuple(2);
  FIELD(res, 0) = MLINT(fd[0]);
  FIELD(res, 1) = MLINT(fd[1]);
  return res;
}

value unix_fork()               /* ML */
{
  return MLINT(fork());
}

value unix_execv(name, args)    /* ML */
     value name, args;
{
  unsigned char ** argv;
  blocksize argc, i;

  argc = SIZE(args);
  argv = (unsigned char **) temp_alloc((argc + 1) * sizeof(char *));
  if (argv == 0) system_failwith_out_of_memory();
  for (i = 0; i < argc; i++)
    argv[i] = CSTRING(FIELD(args, i));
  argv[argc] = 0;
  execv(CSTRING(name), argv);
  temp_free();
  return atom(0);
}

value unix_times()              /* ML */
{
  value res;
  struct tms buffer;
  long real_time;
  int i;

#ifndef HZ
#define HZ 60
#endif

  real_time = times(&buffer);
  *--extern_asp = alloc_double((double) real_time / HZ);
  *--extern_asp = alloc_double((double) buffer.tms_utime / HZ);
  *--extern_asp = alloc_double((double) buffer.tms_stime / HZ);
  *--extern_asp = alloc_double((double) buffer.tms_cutime / HZ);
  *--extern_asp = alloc_double((double) buffer.tms_cstime / HZ);
  res = alloc_tuple(5);
  for (i = 4; i >= 0; i--)
    FIELD(res, i) = *extern_asp++;
  return res;
}

#endif /* ! UNIX */

